/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#define _XOPEN_SOURCE	500

#define NEW_METHOD_FOR_IPV6

#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <limits.h>
#include <time.h>
#include <errno.h>

#include <openssl/err.h>

#include "xacml_client.h"
#include "xacml_server.h"

#include "xacml_io_ssl.h"

#include "xacml_io_ssl_log.h"

/* Network and SSL */
#include "net_common.h"
#include "ssl-common.h"
/* Network and SSL */
#include <stdlib.h>

#ifdef USE_LCAS_ON_SSL_LAYER
#   include <lcas/lcas.h>
#   include <lcas/lcas_pem.h>
#   include <lcas/lcas_types.h>
#endif

#ifdef CLOCK_MONOTONIC_RAW
# define XACML_IO_CLOCK_ID  CLOCK_MONOTONIC_RAW
#else
# define XACML_IO_CLOCK_ID  CLOCK_MONOTONIC
#endif


/* State of this thread */
typedef struct
{
    int             socket;
    SSL            *ssl;
    int             failed;
}
ssl_io_state_t;


/* Prototype */
static void *ssl_io_accept(int msocket,
#ifdef  NEW_METHOD_FOR_IPV6
    struct sockaddr_storage *addr,
#else
    struct sockaddr         *addr,
#endif
    socklen_t *addr_len, int *sock_out);
static void *ssl_io_connect(const char *endpoint, const char *host, int port);
static int ssl_io_send(void *arg, const char *data, size_t size); 
static size_t ssl_io_recv(void *arg, char *data, size_t size); 
static int ssl_io_close(void *arg);

xacml_io_descriptor_t
xacml_io_ssl_descriptor =
{
    "xacml_io_ssl_descriptor",
    (xacml_io_accept_t)ssl_io_accept,
    ssl_io_connect,
    ssl_io_send,
    ssl_io_recv,
    ssl_io_close
};


#ifdef  NEW_METHOD_FOR_IPV6
static
void *
ssl_io_accept(
    int                                 msocket,
    struct sockaddr_storage            *addr,
    socklen_t                          *addr_len,
    int                                *sock_out)
{
/*    int                 rc             = 0;*/
    ssl_io_state_t *    state          = NULL;
    char *              remote_host    = NULL;
    SSL_CTX *           my_ctx         = NULL;
#ifdef USE_LCAS_ON_SSL_LAYER
    long                len            = 0;
    char *              ssl_subject_dn = NULL;
    char *              lcas_log_file_env = NULL;

    BIO *		certbio_mem = NULL;
    STACK_OF (X509) *	px509_chain = NULL;
    X509 *		px509       = NULL;
    char *		pem_string = NULL;
    int			i = 0;

    int			retval = 0;
    char		tmp_dn[256];
    /* Note... the 256 buffer length is only used for logging output on the spot
     * here. More authoritive information is in the LCAS and LCMAPS logs */
#endif


    XACML_LOG_DEBUG(5, "Entering %s\n", __func__);


    /* Allocate thread specific state */
    state = calloc (1,sizeof (ssl_io_state_t));
    if (!state)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: memory allocation of ssl_io_state_t failed\n", __func__);
/*        rc = -1;*/
        return NULL;
    }
    else
    {
	/* initialize socket to -1
	 * ssl and failed are initialized by calloc to 0 */
        state->socket = -1;
        /*
        state->socket = 0;
        state->ssl    = NULL;
        state->close_sock = SOCKET_CAN_CLOSE;
        state->failed = FALSE;
        */
    }



    /******* SSL Init *********/
    my_ctx = xacml_io_ssl_ctx;
    if (!my_ctx)
    {
        my_ctx = xacml_io_setup_server_ctx(NULL,
					   xacml_io_certCAPath,
					   xacml_io_certFilePub,
					   xacml_io_certFilePriv);
        if (!my_ctx)
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Error: unable to create an SSL_CTX context with xacml_io_setup_server_ctx()\n", __func__);
            goto skip_everything_and_exit;
        }
        else
        {
            XACML_LOG_DEBUG(5, "%s: Created CTX here\n", __func__);
        }
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: Got SSL_CTX object\n", __func__);
    }
    /**************************/


    /* Setup of Master Socket to listen on the port */
    if (msocket <= 0)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Failure in master listening socket\n", __func__);
        goto skip_everything_and_exit;
    }



    /* Accept and create client socket */
    if ((xacml_io_accept_new_client_socket (msocket,
                                    &(state -> socket),
                                    &remote_host) ) ||
                                                    (state -> socket < 0) )
    {
        /* Don't overdue the error message, the error is mentioned in more detail in accept_new_client_socket() */
        /* XACML_LOG_FUNC(LOG_ERR, "%s: Failed to accept the client socket connection!\n", __func__); */
/*        rc = 14;*/
        goto finalize_TCPIP;
    }
    else
    {
        XACML_LOG_FUNC(LOG_INFO, "%s: TCP connection accepted from host: %s\n", __func__, remote_host);
        /* SSL stuff */
        if ((state -> ssl = xacml_io_SSL_server_accept (state -> socket,
                                               my_ctx
                                              )) == NULL)
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Failure in SSL layer setup and connection for host: %s\n", __func__, remote_host);
/*            rc = 15;*/
            goto finalize_TCPIP;
        }
        else
        {
            XACML_LOG_FUNC(LOG_DEBUG, "%s: SSL layer (server side) connection accepted from host %s\n", __func__, remote_host);

#ifdef USE_LCAS_ON_SSL_LAYER
	    /* Initialize LCAS - Authorizing SSL client certificate */
	    /* When set, use LCAS_LOG_FILE */
	    if ( (lcas_log_file_env = getenv("LCAS_LOG_FILE"))!=NULL)	{
		if (lcas_init_and_logfile(lcas_log_file_env, NULL, DO_USRLOG))
		{
		    XACML_LOG_FUNC(LOG_ERR, "%s: Initialization of LCAS failed!\n", __func__);
		    goto finalize_SSL;
		}
		else
		    XACML_LOG_DEBUG(5, "LCAS logs to %s\n", lcas_log_file_env);
	    }
	    else /* LCAS_LOG_FILE is unset */
	    {
		if (lcas_init_and_logfile(NULL, scas_get_log_file_pointer(), scas_get_log_type())!=0)	{
		    XACML_LOG_FUNC(LOG_ERR, "%s: Initialization of LCAS failed!\n", __func__);
		    goto finalize_SSL;
		} else
		    XACML_LOG_DEBUG(5, "LCAS logs successfully integrated with log file\n");
	    }

	    /* TODO: I should make a beter and cleaner interface in LCAS to do this */
	    certbio_mem = BIO_new(BIO_s_mem());
	
	    /* Get certificate chain */
	    if (!(px509_chain = SSL_get_peer_cert_chain(state -> ssl)))
	    {
		XACML_LOG_FUNC(LOG_ERR, "%s: Error: Could not get peer chain of certificates from the SSL handle\n", __func__);
		goto finalize_SSL;
	    }

	    /* Push certificates in chain into the BIO memory stack */
	    for (i = 0; i < sk_X509_num(px509_chain); i++)
	    {
		px509 = sk_X509_value(px509_chain, i);
		X509_NAME_oneline(X509_get_subject_name(px509), tmp_dn, 256);
		XACML_LOG_DEBUG(5, "%s: Subject is: %s\n", __func__, tmp_dn);
		retval = PEM_write_bio_X509(certbio_mem, px509);
		XACML_LOG_DEBUG(5, "%s: PEM_write_bio_X509 from stack item %d, %s\n", __func__, i, (retval ? "succeeded" : "failed"));
	    }
	   
	    /* Report the subject DN of the connecting party */
	    ssl_subject_dn = xacml_io_x509_chain_to_dn (px509_chain);
	    XACML_LOG_FUNC(LOG_NOTICE, "%s: \"%s\" is connected from \"%s\"\n", __func__, ssl_subject_dn, remote_host);
	    free (ssl_subject_dn);
	    ssl_subject_dn = NULL;

	    /* Push BIO memory to character array */
	    /* Oscar
	     * Warning: The BIO_get_mem_data() will only apply a REFERENCE to
	     * the pem_string, NOT A COPY of the data
	     */
	    len = BIO_get_mem_data (certbio_mem, &pem_string);
	    pem_string[len]='\0';

	    XACML_LOG_DEBUG(5, "%s: PEM string constructed for LCAS use on SSL layer.\n", __func__);

	    if (lcas_pem(pem_string, "RSL") != 0)
	    {
		XACML_LOG_FUNC(LOG_ERR, "%s: Client credentials are not authorized!\n", __func__);
		if (lcas_term())
		{
		    XACML_LOG_FUNC(LOG_WARNING, "%s: LCAS termination failure.\n", __func__);
		}

		/*
		 * Closing and Freeing buffers AFTER making use of them in LCAS
		 */
		BIO_set_close(certbio_mem, BIO_CLOSE);
		BIO_free(certbio_mem);

		goto finalize_SSL;
	    }
	    else
	    {
		XACML_LOG_DEBUG(3, "%s: Client credentials authorized to send query\n", __func__);

		/* LCAS term */
		if (lcas_term())
		{
		    XACML_LOG_FUNC(LOG_WARNING, "%s: LCAS termination failure.\n", __func__);
		    /*rc = 103;*/
		}

		/*
		 * Closing and Freeing buffers AFTER making use of them in LCAS
		 */
		BIO_set_close(certbio_mem, BIO_CLOSE);
		BIO_free(certbio_mem);
	    }
#endif /* USE_LCAS_ON_SSL_LAYER */

            goto succesful_accept;
        }
        /* SSL stuff */
    }



/*** Good exit ***/
succesful_accept:
    /* Return the connected client socket to the XACML library to listen for traffic */
    free(remote_host);
    *sock_out = state -> socket;
    XACML_LOG_DEBUG(5, "%s: done\n", __func__);
    return state;

/*** Bad exit ***/
#ifdef USE_LCAS_ON_SSL_LAYER
finalize_SSL:
#endif /* USE_LCAS_ON_SSL_LAYER */
#if 0
    if (state && state -> ssl)
    {
        /* Shutdown active SSL layer */
        if (xacml_io_SSL_connect_shutdown (state -> ssl))
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Error in Server xacml_io_SSL_connect_shutdown!\n", __func__);
        }
        SSL_free(state -> ssl);
    }
#endif

finalize_TCPIP:
#if 0
    if (state)
    {
        close (state -> socket);
        state -> socket = 0;
    }
    free(addr);
#endif

skip_everything_and_exit:
    /*
    free(state);
    state = NULL;
    */
    free(remote_host);

    state->failed = 1;

    ssl_io_close (state);
    *sock_out = -1;

    XACML_LOG_DEBUG(2, "%s: ended in a failure\n", __func__);

    /* return NULL; */
    return NULL;
}
#else /* NEW_METHOD_FOR_IPV6 */
static
void *
ssl_io_accept(
    int                                 msocket,
    struct sockaddr                    *addr,
    socklen_t                          *addr_len,
    int                                *sock_out)
{
/*    int              rc = 0;*/
    ssl_io_state_t *    state          = NULL;
    char *              remote_host    = NULL;
    SSL_CTX *           my_ctx         = NULL;
#ifdef USE_LCAS_ON_SSL_LAYER
    long                len            = 0;
    char *              ssl_subject_dn = NULL;
    char *              lcas_log_file_env = NULL;
#endif


    XACML_LOG_DEBUG(5, "Entering %s\n", __func__);

/*
 *   XACML_LOG_DEBUG(5, "Master Socket is: %d.\n", msocket);
 *   if (fcntl (msocket, F_SETFL, O_NONBLOCK) == -1)
 *   {
 *       XACML_LOG_DEBUG(5, "Setting master socket to non-blocking failed\n", __func__);
 *       return NULL;
 *   }
 */


    /* Allocate thread specific state */
    state = calloc (1, sizeof (ssl_io_state_t));
    if (!state)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: memory allocation of ssl_io_state_t failed\n", __func__);
/*        rc = -1;*/
        return NULL;
    }
    else
    {
	/* initialize socket to -1
	 * ssl and failed are initialized by calloc to 0 */
        state->socket = -1;
        /*
        state->socket = 0;
        state->ssl    = NULL;
        state->close_sock = SOCKET_CAN_CLOSE;
        state->failed = FALSE;
        */
    }



    /******* SSL Init *********/
    my_ctx = xacml_io_ssl_ctx;
    if (!my_ctx)
    {
        my_ctx = xacml_io_setup_server_ctx(NULL,
					   xacml_io_certCAPath,
					   xacml_io_certFilePub,
					   xacml_io_certFilePriv);
        if (!my_ctx)
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Error: unable to create an SSL_CTX context with xacml_io_setup_server_ctx()\n", __func__);
            goto skip_everything_and_exit;
        }
        else
        {
            XACML_LOG_DEBUG(5, "%s: Created CTX here\n", __func__);
        }
    }
    else
    {
        XACML_LOG_DEBUG(5, "%s: Got SSL_CTX object\n", __func__);
    }
    /**************************/


    /* Setup of Master Socket to listen on the port */
    if (msocket <= 0)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: Failure in master listening socket\n", __func__);
        goto skip_everything_and_exit;
    }


    *addr_len = sizeof(struct sockaddr_in);


    /* Accept and create client socket */
    if ((xacml_io_accept_new_client_socket (msocket,
					    &(state -> socket),
					    &remote_host) ) ||
        (state -> socket < 0) )
    {
        /* Don't overdue the error message, the error is mentioned in more detail in accept_new_client_socket() */
        /* XACML_LOG_FUNC(LOG_ERR, "%s: Failed to accept the client socket connection!\n", __func__); */
/*        rc = 14;*/
        goto finalize_TCPIP;
    }
    else
    {
        XACML_LOG_FUNC(LOG_INFO, "%s: TCP connection accepted from host: %s\n", __func__, remote_host);
        /* SSL stuff */
        if ((state -> ssl = xacml_io_SSL_server_accept (state -> socket,
                                               my_ctx
                                      )) == NULL)
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Failure in SSL layer setup and connection for host: %s\n", __func__, remote_host);
/*            rc = 15;*/
            goto finalize_TCPIP;
        }
        else
        {
            XACML_LOG_FUNC(LOG_INFO, "%s: SSL layer (server side) connection accepted from host %s\n", __func__, remote_host);

#ifdef USE_LCAS_ON_SSL_LAYER
            /* Initialize LCAS - Authorizing SSL client certificate */
	    /* When set, use LCAS_LOG_FILE */
	    if ( (lcas_log_file_env = getenv("LCAS_LOG_FILE"))!=NULL)	{
		if (lcas_init_and_logfile(lcas_log_file_env, NULL, DO_USRLOG))
		{
		    XACML_LOG_FUNC(LOG_ERR, "%s: Initialization of LCAS failed!\n", __func__);
		    goto finalize_SSL;
		}
		else
		    XACML_LOG_DEBUG(5, "LCAS logs to %s\n", lcas_log_file_env);
	    }
	    else /* LCAS_LOG_FILE is unset */
	    {
		if (lcas_init_and_logfile(NULL, XACML_GET_LOG_FILE_POINTER(), XACML_GET_LOG_TYPE())!=0)	{
		    XACML_LOG_FUNC(LOG_ERR, "%s: Initialization of LCAS failed!\n", __func__);
		    goto finalize_SSL;
		} else
		    XACML_LOG_DEBUG(5, "LCAS logs successfully integrated with log file\n");
	    }

	    /* TODO: I should make a beter and cleaner interface in LCAS to do this */
	    BIO * certbio_mem = NULL;
	    STACK_OF (X509) * px509_chain = NULL;
	    X509 *            px509       = NULL;
	    char * pem_string = NULL;
	    int i = 0;

	    certbio_mem = BIO_new(BIO_s_mem());
	
	    /* Get certificate chain */
	    if (!(px509_chain = SSL_get_peer_cert_chain(state -> ssl)))
	    {
		XACML_LOG_FUNC(LOG_ERR, "%s: Error: Could not get peer chain of certificates from the SSL handle\n", __func__);
		goto finalize_SSL;
	    }

	    /* Push certificates in chain into the BIO memory stack */
	    for (i = 0; i < sk_X509_num(px509_chain); i++)
	    {
		int retval = 0;
		char tmp_dn[256]; /* Note... the 256 buffer length is only used for logging output on the spot here. More authoritive information is in the LCAS and LCMAPS logs */

		px509 = sk_X509_value(px509_chain, i);
		X509_NAME_oneline(X509_get_subject_name(px509), tmp_dn, 256);
		XACML_LOG_DEBUG(5, "%s: Subject is: %s\n", __func__, tmp_dn);
		retval = PEM_write_bio_X509(certbio_mem, px509);
		XACML_LOG_DEBUG(5, "%s: PEM_write_bio_X509 from stack item %d, %s\n", __func__, i, (retval ? "succeeded" : "failed"));
	    }
	   
	    /* Report the subject DN of the connecting party */
	    ssl_subject_dn = xacml_io_x509_chain_to_dn (px509_chain);
	    XACML_LOG_FUNC(LOG_ERR, "%s: \"%s\" is connected from \"%s\"\n", __func__, ssl_subject_dn, remote_host);
	    free (ssl_subject_dn);
	    ssl_subject_dn = NULL;

	    /* Push BIO memory to character array */
	    /* Oscar
	     * Warning: The BIO_get_mem_data() will only apply a REFERENCE to
	     * the pem_string, NOT A COPY of the data
	     */
	    len = BIO_get_mem_data (certbio_mem, &pem_string);
	    pem_string[len]='\0';

	    XACML_LOG_DEBUG(5, "%s: PEM string constructed for LCAS use on SSL layer.\n", __func__);

	    if (lcas_pem(pem_string, "RSL") != 0)
	    {
		XACML_LOG_FUNC(LOG_ERR, "%s: Client credentials are not authorized!\n", __func__);
		if (lcas_term())
		{
		    XACML_LOG_FUNC(LOG_WARNING, "%s: LCAS termination failure.\n", __func__);
		}

		/*
		 * Closing and Freeing buffers AFTER making use of them in LCAS
		 */
		BIO_set_close(certbio_mem, BIO_CLOSE);
		BIO_free(certbio_mem);

		goto finalize_SSL;
	    }
	    else
	    {
		XACML_LOG_DEBUG(3, "%s: Client credentials authorized to send query\n", __func__);

		/* LCAS term */
		if (lcas_term())
		{
		    XACML_LOG_FUNC(LOG_WARNING, "%s: LCAS termination failure.\n", __func__);
		    /* rc = 103;*/
		}

		/*
		 * Closing and Freeing buffers AFTER making use of them in LCAS
		 */
		BIO_set_close(certbio_mem, BIO_CLOSE);
		BIO_free(certbio_mem);
	    }
#endif /* USE_LCAS_ON_SSL_LAYER */
            goto succesful_accept;
        }
        /* SSL stuff */
    }


/*** Good exit ***/
succesful_accept:
    /* Return the connected client socket to the XACML library to listen for traffic */
    if (remote_host) free(remote_host);
    *sock_out = state -> socket;
    XACML_LOG_DEBUG(5, "%s: done\n", __func__);
    return state;


/*** Bad exit ***/
/* finalize_SSL: */
#if 0
    if (state && state -> ssl)
    {
        /* Shutdown active SSL layer */
        if (xacml_io_SSL_connect_shutdown (state -> ssl))
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Error in Server xacml_io_SSL_connect_shutdown!\n", __func__);
        }
        SSL_free(state -> ssl);
    }
#endif

finalize_TCPIP:
#if 0
    if (state)
    {
        close (state -> socket);
        state -> socket = 0;
    }
    free(addr);
#endif

skip_everything_and_exit:
    /*
    free(state);
    state = NULL;
    */
    free(remote_host);

    state->failed = 1;

    ssl_io_close (state);
    *sock_out = -1;

    XACML_LOG_DEBUG(2, "%s: ended in a failure\n", __func__);

    return NULL;
}
#endif /* NEW_METHOD_FOR_IPV6 */




static
void *
ssl_io_connect(
    const char                         *endpoint,
    const char                         *host,
    int                                 port)
{
#ifdef OVERRIDE_API_HOST_CHECK
    char            *override_hostname         = NULL;
#endif
    ssl_io_state_t  *state = NULL;

#if 1
    /* Initialized with the old settings */
    long             realtime_timeout      = 1; /* seconds */
    long	     backoff_milli_seconds = 50;  /* milli seconds */
    long             socket_timeout        = 170; /* milli seconds */

    realtime_timeout      = xacml_io_get_connection_timeout();   /* seconds */
    backoff_milli_seconds = xacml_io_get_incremental_backoff();  /* milli seconds */
    socket_timeout        = xacml_io_get_socket_timeout();       /* milli seconds */
#else
    long             realtime_timeout      = 120;      /* seconds */
    long	     backoff_milli_seconds = 200;     /* milli seconds */
    long             socket_timeout        = 30000;   /* milli seconds */
#endif

    long             err     = 0;
/*    int              rc      = 0;*/
    int              conn_rc = 0;
/*    time_t           start   = 0;*/
    struct timespec  t1,t2,backoff;
    unsigned int     seed    = 0;


    XACML_LOG_DEBUG(5, "Entering %s\n", __func__);

    /* Initialize OpenSSL Library, as it might not have been initialized before */
    xacml_io_init_OpenSSL();

    /* Random function seeding */
    seed = (unsigned int)time((time_t *)NULL);
    srand(seed);
    /* Random function seeding */

    XACML_LOG_DEBUG(5, "%s: realtime_timeout (%d), backoff_milli_seconds (%ld), socket_timeout (%d)\n",
                        __func__, realtime_timeout, backoff_milli_seconds, socket_timeout);

    /* Allocate thread specific state */
    state = calloc (1,sizeof (ssl_io_state_t));
    if (!state)
    {
        XACML_LOG_FUNC(LOG_ERR, "%s: memory allocation of ssl_io_state_t failed\n", __func__);
/*        rc = -1;*/
        return NULL;
    }
    else
    {
	/* initialize socket to -1
	 * ssl and failed are initialized by calloc to 0 */
        state->socket = -1;
    }


    if (clock_gettime(XACML_IO_CLOCK_ID,&t1))	{
        XACML_LOG_FUNC(LOG_ERR, "%s: cannot get time of day: %s\n",
		__func__, strerror(errno));
	t1.tv_sec=0; t1.tv_nsec=0;
    }

/*    start = time(NULL);*/
/*    while ((time(NULL) - start) < realtime_timeout)*/
    do
    {
        /* Create client socket */
        XACML_LOG_DEBUG(5, "%s: Connecting socket to host %s on port %d\n", __func__, host, port);
        conn_rc = xacml_io_create_client_socket (&(state -> socket),
						 host,
						 port,
						 socket_timeout);
        switch (conn_rc)
        {
            case 2 : /* getaddrinfo() failed to resolve DNS name to IP */
                goto finalize_TCPIP;
                break;
            case 1 : /* Generic failure */
            case 3 : /* Failed to created socket() */
            case 4 : /* Failed to get timeout setting on the socket */
            case 5 : /* Failed to set timeout setting on the socket */
            case 6 : /* connect() failed */
                /* Retry connection setup, with a random wait */
                XACML_LOG_DEBUG(2, "%s: Socket creation failed, retrying... in %ld msec!\n", __func__, backoff_milli_seconds);

                /* Exponential back off */
		backoff.tv_sec=backoff_milli_seconds/1000L;
		backoff.tv_nsec=(backoff_milli_seconds-backoff.tv_sec*1000L)*1000000L;
/*                usleep (backoff_milli_seconds * 1000);*/
		nanosleep(&backoff,NULL);
                backoff_milli_seconds *= 2;
                backoff_milli_seconds += (unsigned)(rand_r (&seed)) % 25;

		if (clock_gettime(XACML_IO_CLOCK_ID,&t2)) {
		    XACML_LOG_FUNC(LOG_ERR, "%s: cannot get time of day: %s\n",
			    __func__, strerror(errno));
		    break;
		}

                continue; /* Continue until the timeout is reached */
        }

        /* TCP/IP fully connected */
        XACML_LOG_DEBUG(5, "%s: Connected on TCP/IP\n", __func__);
        XACML_LOG_FUNC(LOG_INFO, "%s: Is connected to host '%s' on port '%d'\n", __func__, host, port);

        /* SSL */
        if ((state -> ssl = xacml_io_SSL_client_connect (state -> socket,
                                                xacml_io_setup_client_ctx (NULL,
                                                                  xacml_io_certCAPath,
                                                                  xacml_io_certFilePub,
                                                                  xacml_io_certFilePriv)
                                      )) == NULL)
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Failure in SSL layer setup and connection!\n", __func__);

            /* Finalize TCP/IP */
/*            rc = 15;*/
            goto finalize_TCPIP;
        }
        else
        {
            XACML_LOG_DEBUG(5, "%s: Connected on SSL!\n", __func__);

	    /* Post SSL connection check */
#ifdef OVERRIDE_API_HOST_CHECK
            /* Check if there is a hostname override set, if not, perform regular checks */
            if ((override_hostname = xacml_io_get_hostname_to_match_client_cert()))
            {
                XACML_LOG_DEBUG(5, "%s: Overriding hostname check from \'%s\' to \'%s\'\n", __func__, host, override_hostname);
                err = xacml_io_post_connection_check(state -> ssl, override_hostname);

                /* Purge allocated memory, as it has served its purpose */
/*                xacml_io_free_hostname_to_match_client_cert();*/
            }
            else
            {
                err = xacml_io_post_connection_check(state -> ssl, host);
            }
#else
            /* No override possible, the SSL peer certificate must match the 'host' value */
            err = xacml_io_post_connection_check(state -> ssl, host);
#endif
            if (err != X509_V_OK)
            {
                XACML_LOG_FUNC(LOG_ERR, "%s: Error with peer certificate after the full handshake: %s\n", __func__, X509_verify_cert_error_string(err));
                /* SSL_set_verify_result (state -> ssl, err);*/

                /* Finalize SSL connection */
/*                rc = 16;*/
                goto finalize_SSL;
            }
            else
            {
                XACML_LOG_DEBUG(5, "%s: SSL layer (client side) connection done\n", __func__);

                /* Finalize SSL connection */
/*                rc = 0;*/
                goto succesful_connect;
	    }
        }
        /* SSL */

	if (clock_gettime(XACML_IO_CLOCK_ID,&t2)) {
	    XACML_LOG_FUNC(LOG_ERR, "%s: cannot get time of day: %s\n",
		    __func__, strerror(errno));
	    break;
	}

    } while ( (double)(t2.tv_sec-t1.tv_sec)+((double)(t2.tv_nsec-t1.tv_nsec))/1e9 < realtime_timeout);
/*    while ((time(NULL) - start) < realtime_timeout)*/

    /* Failure */
    XACML_LOG_FUNC(LOG_INFO, "%s: Failed to connect to host '%s' on port '%d'\n", __func__, host, port);
    goto finalize_TCPIP;


/* Good exit */
succesful_connect:
    XACML_LOG_DEBUG(5, "%s: done\n", __func__);
    return state;


/* Bad exit */
finalize_SSL:
    ssl_io_close (state);
    XACML_LOG_DEBUG(3, "%s: ended in a failure\n", __func__);
    return NULL;

    /* There should be an SSL object to destroy in the first place */
#if 0
    if (state && state -> ssl)
    {
        /* Shutdown active SSL layer */
        if (xacml_io_SSL_connect_shutdown (state -> ssl))
        {
            XACML_LOG_FUNC(LOG_ERR, "%s: Error: in client xacml_io_SSL_connect_shutdown!\n", __func__);
        }
        SSL_free(state -> ssl);
    }
#endif

finalize_TCPIP:
    if (state)
    {
        close (state -> socket);
        free(state);
        state = NULL;
    }

    XACML_LOG_FUNC(LOG_INFO, "%s: ended in a failure\n", __func__);
    return NULL;
}
/* ssl_io_connect() */


static
int
ssl_io_send(
    void                               *arg,
    const char                         *data,
    size_t                              size)
{
    ssl_io_state_t *                    state = arg;
    int                                 rc = -2;
    int                                 nwritten = 0;

    XACML_LOG_DEBUG(5, "%s: Sending...\n", __func__);

    /* Check whether size fits in an int since SSL_write() uses int */
    if (size>INT_MAX)   {
	XACML_LOG_FUNC(LOG_ERR,
		"%s: cannot pass size %ld to SSL_write() which expects an int.\n",
		__func__, (long unsigned)size );
	return -1;
    }

    if (state && state -> ssl)
    {
	/* Note: we're only adding rc when it is >0 */
        for (nwritten = 0; nwritten < (int)size; nwritten += rc)
        {
            rc = SSL_write (state -> ssl, data + nwritten, (int)size - nwritten);
            /* Error conditions are either zero or < zero. */
            if (rc <= 0)
            {
                xacml_io_print_ssl_error_msg (state -> ssl, rc, 1, NULL);
                return -1;
            }
        }
    }

    if (BIO_flush(state -> ssl -> wbio))
        XACML_LOG_DEBUG(5, "%s: BIO (write buffer) flushed\n", __func__);

    XACML_LOG_DEBUG(5, "%s: Send %d of %d\n", __func__, rc, size);
    return 0;
}
/* ssl_io_send() */


/* Ultimately used as soap.frecv, see for frecv() in stdsoap2.c
 * There it seems frecv should return 0 on error or num of bytes read on
 * success. */
static
size_t
ssl_io_recv(
    void                               *arg,
    char                               *data,
    size_t                              size)
{
    ssl_io_state_t * state = arg;
    int              rc = 0;


    if (state && state -> ssl)
    {
	if (size>INT_MAX)   {
	    XACML_LOG_FUNC(LOG_ERR,
		    "%s: cannot pass size %ld to xacml_io_SSL_recv() which expects an int.\n",
		    __func__, (long unsigned)size );
	    /* See above: 0 on error */
	    return 0;
	}

        if ((rc = xacml_io_SSL_recv (state -> ssl, data, (int)size)) < 0)
        {
            if (BIO_flush(state -> ssl -> rbio))
                XACML_LOG_DEBUG(5, "%s: BIO (read buffer) flushed\n", __func__);

            XACML_LOG_FUNC(LOG_ERR, "%s: Failure in xacml_io_SSL_recv\n", __func__);
	    /* See above: 0 on error */
            return 0;
        }
    }

    XACML_LOG_DEBUG(5, "%s: End of ssl_io_recv()\n", __func__);
    /* rc >=0 */
    return (size_t)rc;
}
/* ssl_io_recv() */



static
int
ssl_io_close(
             void  *arg)
{
    ssl_io_state_t *                state = arg;
    int                             failed_state = 0;

    XACML_LOG_DEBUG(5, "%s: Entering ssl_io_close()\n", __func__);

    /* Only work with it, if not already destroyed or set to NULL */
    if (state)
    {
        failed_state = state -> failed;

        if (state -> ssl)
        {
            /* Shutdown active SSL layer */
            if (xacml_io_SSL_connect_shutdown (state -> ssl))
            {
                XACML_LOG_FUNC(LOG_NOTICE, "%s: Error: in xacml_io_SSL_connect_shutdown!\n", __func__);
            }
            state -> ssl -> rbio = NULL;
            state -> ssl -> wbio = NULL;
            XACML_LOG_DEBUG(3, "%s: going to SSL_free()\n", __func__);

            SSL_free(state -> ssl);
            state -> ssl = NULL;
        }

	/* Close the socket */
	if (state -> socket>=0)
	    close(state->socket);

	/* Free the state */
	free(state);
	state = NULL;

    }
    
    /* Has to remove the error state here, because the thread is about to end,
     * otherwise we can no longer free the memory */
    ERR_remove_state(0);

    XACML_LOG_DEBUG(5, "%s: ended perfectly.\n", __func__);
    return failed_state;
}
/* ssl_io_close() */

